from datetime import datetime, timedelta


def get_date(datestring: str = "") -> datetime:
    try:
        date = datetime.strptime(datestring, "%Y-%m-%d")
    except:
        date = datetime.today()
    finally:
        return date


def delta_week(date: datetime, delta: int = 0) -> datetime:
    return date + timedelta(weeks=delta)


def get_year(datestring: str = "") -> int:
    return delta_week(get_date(datestring), -33).strftime("%Y")



import aiohttp
import logging

logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)


async def get_json(url: str):
    async with aiohttp.ClientSession() as client:
        async with client.get(url) as response:
            return await response.read()


async def post_request(url: str, data: dict) -> str:
    async with aiohttp.ClientSession() as session:
        response = await session.post(url=url, data=data)
        result = await response.json()
        if "access_token" in result:
            return result["access_token"]
        return ""



from ._zermelo_collection import ZermeloCollection, from_zermelo_dict
from dataclasses import dataclass, InitVar, field
import logging

logger = logging.getLogger(__name__)


@dataclass
class User:
    code: str
    roles: list[str]
    firstName: str
    prefix: str
    lastName: str
    schoolInSchoolYears: list[int]
    isApplicationManager: bool
    archived: bool
    isStudent: bool
    isEmployee: bool
    isFamilyMember: bool
    hasPassword: bool
    isSchoolScheduler: bool
    isSchoolLeader: bool
    isStudentAdministrator: bool
    isTeamLeader: bool
    isSectionLeader: bool
    isMentor: bool
    isParentTeacherNightScheduler: bool
    isDean: bool
    fullName: str = ""

    def __post_init__(self):
        self.fullName = self.generatename()

    def __repr__(self):
        return f"{self.fullName} ({self.code})"

    def generatename(self) -> str:
        if self.prefix:
            fullname = " ".join(
                [self.firstName, self.prefix, self.lastName.split(",")[0]]
            )
        elif self.firstName is not None and self.lastName is not None:
            fullname = " ".join([self.firstName, self.lastName])
        elif self.lastName is not None:
            fullname = self.lastName
        else:
            fullname = "unknown"
        return fullname


@dataclass
class LeerjaarCounter:
    id: int
    count: int = 0


class LeerjaarCounters(list[LeerjaarCounter]):
    def get(self, id) -> LeerjaarCounter:
        for ljcounter in self:
            if ljcounter.id == id:
                return ljcounter
        self.append(LeerjaarCounter(id))
        return self[-1]

    def add(self, leerjaar_id: int):
        counter = self.get(leerjaar_id)
        counter.count += 1

    def get_id(self) -> int:
        if len(self):
            self.sort(key=lambda x: x.count, reverse=True)
            return self[0].id
        return 0


@dataclass
class Leerling(User):
    volgnr: int = 0
    leerjaren: LeerjaarCounters = field(default_factory=LeerjaarCounters)


class Leerlingen(ZermeloCollection[Leerling]):

    def __init__(self, schoolinschoolyear: int = 0):
        query = f"users?schoolInSchoolYear={schoolinschoolyear}&isStudent=true"
        super().__init__(Leerling, query)

    async def _init(self):
        data = await self.get_collection()
        if data:
            self.load_leerlingen(data)

    def load_leerlingen(self, data: list[dict]):
        logger.info("loading Leerlingen")
        data.sort(key=lambda x: (x["lastName"], x["firstName"]))
        for idx, row in enumerate(data):
            self.append(from_zermelo_dict(Leerling, row, volgnr=idx + 1))
        logger.info(f"found: {len(self)} leerlingen")

    def __repr__(self):
        return f"{self}{self.print_list()}"

    def __str__(self):
        return f"Leerlingen({len(self)})"

    def get(self, llnr) -> Leerling | None:
        for user in self:
            if user.code == str(llnr):
                return user


@dataclass
class Medewerker(User):
    def __repr__(self):
        return self.fullName + f"({self.code})"


class Personeel(ZermeloCollection[Medewerker]):

    def __init__(self, schoolinschoolyear: int = 0):
        query = f"users?schoolInSchoolYear={schoolinschoolyear}&isEmployee=true"
        super().__init__(Medewerker, query)

    def __repr__(self):
        return f"{self}{self.print_list()}"

    def __str__(self):
        return f"Personeel({len(self)})"

    def get(self, code: str) -> Medewerker | None:
        for user in self:
            if user.code == code:
                return user



from ._zermelo_collection import ZermeloCollection
from dataclasses import dataclass, InitVar, field
import logging

logger = logging.getLogger(__name__)


@dataclass
class Vak:
    id: int
    subject: int
    departmentOfBranch: int
    studentCanEdit: bool
    sectionOfBranch: int
    courseType: str
    lessonHoursInClassPeriods: list[dict]
    scroungeSegments: list[int]
    excludedSegments: list[int]
    referenceWeek: dict  # (year:int, weekNumber: int, schoolYear: int)
    isExam: bool
    scheduleCode: str
    subjectType: str
    subjectCode: str
    departmentOfBranchCode: str
    iltCode: int
    qualifiedCode: str
    subjectScheduleCode: str
    subjectName: str
    sectionOfBranchAbbreviation: str

    def getName(self) -> str:
        if "/" in self.subjectName:
            logger.debug(f"old name: {self.subjectName}")
            parts = self.subjectName.split("/")
            frontpart = parts[0]
            nameparts = frontpart.split(" ")
            nameparts.pop(-1)
            name = " ".join(nameparts)
            logger.debug(f"new name: {name}")
            return name.strip()
        return self.subjectName.strip()


class Vakken(ZermeloCollection[Vak]):

    def __init__(self, schoolinschoolyear: int = 0):
        query = f"choosableindepartments?schoolInSchoolYear={schoolinschoolyear}"
        super().__init__(Vak, query)

    def get(self, vaknaam: str) -> Vak | None:
        for vak in self:
            if vak.subjectCode == vaknaam:
                return vak

    def get_subject(self, subject: str) -> tuple[int, str]:
        """returns (code, naam)"""
        for vak in self:
            if vak.subjectCode == subject:
                return (vak.subject, vak.getName())
        return (0, "Onbekend")

    def get_leerjaar_vakken(self, leerjaar_id: int, skip: bool = False) -> list[Vak]:
        return [
            vak
            for vak in self
            if vak.departmentOfBranch == leerjaar_id
            and not (
                skip
                and (
                    vak.subjectType in ["education", "profile"]
                    or vak.scheduleCode
                    in [
                        "lo",
                        "sport",
                    ]
                )
            )
        ]



from ._zermelo_collection import ZermeloCollection
from dataclasses import dataclass, InitVar, field
import logging

logger = logging.getLogger(__name__)

@dataclass
class Lokaal:
    id: int
    name: str
    parentteachernightCapacity: int
    courseCapacity: int
    supportsConcurrentAppointments: bool
    allowMeetings: bool
    branchOfSchool: int
    secondaryBranches: list[int]
    schoolInSchoolYear: int


class Lokalen(ZermeloCollection[Lokaal]):

    def __init__(self, schoolinschoolyear: int = 0):
        query = f"locationofbranches?schoolInSchoolYear={schoolinschoolyear}"
        super().__init__(Lokaal, query)

    def get(self, id: int) -> Lokaal | None:
        for lokaal in self:
            if lokaal.id == id:
                return lokaal



from ._config import getConfig
import logging

logger = logging.getLogger(__name__)


class Credentials:
    def __init__(self):
        self.file = getConfig("creds")
        self.token = ""
        self.load()

    def load(self):
        self.token = ""
        try:
            data = self.file.load()
            self.token = data["token"]
        except Exception as e:
            logger.error(e)
        finally:
            logger.debug(f"token: {self.token}")
            return True if self.token else False

    def settoken(self, token):
        self.file.save({"token": token})
        self.token = token



from os import path, mkdir
from shutil import copyfile
import json


class getConfig:
    def __init__(self, file):
        HOMEDIR = getConfigDir()
        filename = file + ".json"
        backupname = file + ".bak"
        if not HOMEDIR:
            raise Exception("Kan configuratie directory niet aanmaken")
        self.filepath = path.join(HOMEDIR, filename)
        self.backup = path.join(HOMEDIR, backupname)

    def __repr__(self):
        return f'ConfigFile(path="{self.filepath}")'

    def save(self, data):
        try:
            copyfile(self.filepath, self.backup)
        except Exception:
            print("warning - geen datafile")
        try:
            with open(self.filepath, "w") as file:
                json.dump(data, file, indent=2)
        except Exception:
            return False
        return True

    def load(self):
        data = {}
        for file in [self.filepath, self.backup]:
            try:
                if path.isfile(file):
                    with open(file, "r") as json_file:
                        data = json.load(json_file)
            except Exception:
                continue
            return data

    def get(self, key: str = ""):
        data = self.load()
        if not key:
            return data
        return data[key] if type(data) is dict and key in data else {}


def getConfigDir():
    abspath = path.abspath(__file__)
    dname = path.dirname(abspath)
    return dname



from ._zermelo_collection import ZermeloCollection
from dataclasses import dataclass, InitVar, field


@dataclass
class VakLokData:
    locationsOfBranch: list[int]
    choosableInDepartments: list[str]


class DataVakLoks(ZermeloCollection[VakLokData]):
    def __init__(self, id_branch: int, start: int, eind: int):
        query = f"appointments?branchOfSchool={id_branch}&fields=locationsOfBranch,choosableInDepartments,&start={start}&end={eind}"
        super().__init__(VakLokData, query)


@dataclass
class VakLok:
    id: int
    lokalen: list[int] = field(default_factory=list)

    def add_loks(self, lokalen: list[int]):
        for lok in lokalen:
            if lok not in self.lokalen:
                self.lokalen.append(lok)


@dataclass
class VakLoks(list[VakLok]):
    def add(self, id) -> VakLok:
        vakdoclok = self.get(id)
        if not vakdoclok:
            vakdoclok = VakLok(id)
            self.append(vakdoclok)
        return vakdoclok

    def get(self, id: int) -> VakLok | None:
        for vakdoclok in self:
            if vakdoclok.id == id:
                return vakdoclok


async def get_vakloks(id_branch: int, start: int, eind: int):
    vakdata = DataVakLoks(id_branch, start, eind)
    await vakdata._init()
    vakdocloks = VakLoks()
    for data in vakdata:
        for id in data.choosableInDepartments:
            vakdoclok = vakdocloks.add(id)
            vakdoclok.add_loks(data.locationsOfBranch)
    return vakdocloks



from ._zermelo_api import zermelo
from typing import TypeVar, Type
import inspect
import logging

logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)

# Define a TypeVar for the specific type of object we want to create
_T = TypeVar("_T")


def from_zermelo_dict(obj: Type[_T], data: dict, *args, **kwargs) -> _T:
    """
    Generates an object of type obj from a dictionary,
    passing only keys that match obj's constructor parameters.

    Args:
        obj: The class (type) to instantiate.
        data: The dictionary containing data for the object's constructor.
        *args: Positional arguments to pass to the constructor.
        **kwargs: Keyword arguments to pass to the constructor.

    Returns:
        An instance of type obj.
    """

    # Get the constructor parameters of the class
    obj_signature = inspect.signature(obj).parameters
    for k, v in data.items():
        if k not in obj_signature:
            logger.debug(f"{k} ({v}) not defined in {obj.__name__}")

    # Filter data to include only keys that match constructor parameters
    filtered_data = {k: v for k, v in data.items() if k in obj_signature}

    # Instantiate the object
    return obj(*args, **filtered_data, **kwargs)


class ZermeloCollection[T](list[T]):

    def __init__(self, item_type: Type[T], query: str = ""):
        super().__init__()
        self.type = item_type
        self.query = query

    def __repr__(self):
        return ", ".join([f"{item!r}" for item in self])

    def print_list(self):
        return (
            f" = [" + ", ".join([str(item) for item in self]) + "]" if len(self) else ""
        )

    async def get_collection(self, query: str = "") -> list[dict]:
        query = self.get_query(query)
        logger.debug(f"type: {self.type}")
        return await zermelo.load_query(query)

    async def load_collection(self, query: str, *args, **kwargs):
        for row in await self.get_collection(query):
            self.append(from_zermelo_dict(self.type, row, *args, **kwargs))

    def get_query(self, query: str) -> str:
        if not query:
            if not self.query:
                raise Exception(f"No query given")
            return self.query
        return query

    async def _init(self, query: str = ""):
        if not self.type:
            raise Exception(f"No Type set for {self}")
        query = self.get_query(query)
        await self.load_collection(query)



from ._zermelo_api import loadAPI
from ._zermelo_collection import ZermeloCollection
from ._time_utils import get_year
from dataclasses import dataclass, field
import logging

logger = logging.getLogger(__name__)


@dataclass
class SchoolInSchoolYear:
    id: int
    name: str
    school: int = field(repr=False)
    year: int = field(repr=False)
    archived: bool = field(repr=False)
    projectName: str = field(repr=False)
    schoolName: str = field(repr=False)
    schoolHrmsCode: str = field(repr=False)


class SchoolYears(ZermeloCollection[SchoolInSchoolYear]):

    def __init__(self, datestring: str = ""):
        year = get_year(datestring)
        logger.debug(year)
        query = f"schoolsinschoolyears/?year={year}&archived=False"
        super().__init__(SchoolInSchoolYear, query)

    def __repr__(self):
        return f"SchoolYears([{super().__repr__()}])"


async def load_schoolyears(schoolname, date: str = "") -> SchoolYears | None:
    try:
        await loadAPI(schoolname)
        schoolyears = SchoolYears(date)
        await schoolyears._init()
        return schoolyears
    except Exception as e:
        logger.error(e)



from ._zermelo_collection import ZermeloCollection
from dataclasses import dataclass, InitVar
import logging

logger = logging.getLogger(__name__)


def getLeerjaarNaam(string):
    names = {"LG1": "klas 1", "LG2": "klas 2", "LG3": "klas 3"}
    for name, result in names.items():
        if name in string:
            return result
    else:
        return string


@dataclass
class Leerjaar:
    id: int
    code: str
    yearOfEducation: int
    branchOfSchool: int
    teacherTeams: list[int]
    mainGroupAuthority: str
    educationType: str
    weekTimeTable: int
    teachingLevel: str
    educations: list[int]
    schoolInSchoolYear: int
    branchOfSchoolCode: str
    schoolInSchoolYearName: str
    profiles: list[int]
    schoolInSchoolYearId: int
    name: str = ""

    def __post_init__(self):
        self.name = getLeerjaarNaam(self.code.upper())


class Leerjaren(ZermeloCollection[Leerjaar]):

    def __init__(self, schoolinschoolyear: int = 0):
        query = f"departmentsofbranches?schoolInSchoolYear={schoolinschoolyear}"
        super().__init__(Leerjaar, query)



from dataclasses import dataclass, field, InitVar
from .vakken import Vakken, Vak
from .vaklessen import get_groep_data, LesData
from .groepen import Groepen, Groep
from .users import Leerlingen, Leerling, Personeel, Medewerker
from .leerjaren import Leerjaren, Leerjaar
from ._zermelo_collection import from_zermelo_dict
import asyncio
import logging

logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)


def createLesgroepNaam(vak: Vak, groep: Groep) -> str:
    leerjaar, groepnaam = groep.extendedName.split(".")
    jaarnaam = leerjaar[2:].upper()
    if vak.subjectCode in groepnaam:
        return f"{jaarnaam}{groepnaam}"
    else:
        return f"{jaarnaam}{vak.subjectCode}{groepnaam[-1]}"


def get_info(
    llnrs: list[int],
    doc_codes: list[str],
    names: list[str],
    lln: Leerlingen,
    docs: Personeel,
) -> tuple[list[Leerling], list[Medewerker], list[str]]:
    leerlingen = [lln.get(llnr) for llnr in llnrs]
    leerlingen = [ll for ll in leerlingen if ll is not None]
    docenten = [docs.get(code) for code in doc_codes]
    docenten = [doc for doc in docenten if doc is not None]
    return (leerlingen, docenten, names)


@dataclass
class Lesgroep:
    vak: Vak
    groep: Groep
    leerjaar: Leerjaar
    leerlingen: list[Leerling] = field(default_factory=list)
    docenten: list[Medewerker] = field(default_factory=list)
    namen: list[str] = field(default_factory=list)
    naam: str = ""

    def __post_init__(self):
        self.naam = createLesgroepNaam(self.vak, self.groep)
        for leerling in self.leerlingen:
            leerling.leerjaren.add(self.leerjaar.id)


class Lesgroepen(list[Lesgroep]):

    @classmethod
    async def create(
        cls,
        leerjaren: Leerjaren,
        vakken: Vakken,
        groepen: Groepen,
        leerlingen: Leerlingen,
        personeel: Personeel,
    ):
        self = cls()
        for leerjaar in leerjaren:
            tasks1 = []
            logger.info(f"finding lesgroepen for {leerjaar.code}")
            for vak in vakken.get_leerjaar_vakken(leerjaar.id, skip=True):
                logger.debug(vak)
                vakgroepen = groepen.get_vakgroepen(vak)
                tasks1.append(
                    asyncio.create_task(
                        find_lesgroepen(
                            leerjaar, vak, vakgroepen, leerlingen, personeel
                        )
                    )
                )
            results: list[tuple[Vak, list[Lesgroep]]] = await asyncio.gather(*tasks1)
            tasks2 = []
            for vak, lesgroepen in results:
                self.extend(lesgroepen)
                if not len(lesgroepen):
                    logger.debug(f"no regular groups found for {vak.subjectName}")
                    maingroepen = groepen.get_department_groups(leerjaar.id, True)
                    tasks2.append(
                        asyncio.create_task(
                            find_lesgroepen(
                                leerjaar, vak, maingroepen, leerlingen, personeel
                            )
                        )
                    )
            if not len(tasks2):
                logger.debug("alle groepen gevonden.")
                continue
            results: list[tuple[Vak, list[Lesgroep]]] = await asyncio.gather(*tasks2)
            for vak, lesgroepen in results:
                self.extend(lesgroepen)
                if not len(lesgroepen):
                    logger.debug(f"geen groepen gevonden voor {vak}")
        self.clean_leerlingen()
        return self

    def clean_leerlingen(self):
        for lesgroep in self:
            for leerling in lesgroep.leerlingen.copy():
                if leerling.leerjaren.get_id() != lesgroep.leerjaar.id:
                    logger.debug(
                        f"removing leerling ({leerling.fullName}) from {lesgroep.naam}"
                    )
                    lesgroep.leerlingen.remove(leerling)


async def find_lesgroepen(
    lj: Leerjaar,
    vak: Vak,
    grpn: list[Groep],
    lln: Leerlingen,
    docs: Personeel,
) -> tuple[Vak, list[Lesgroep]]:
    datalist = await asyncio.gather(*[get_groep_data(vak, groep) for groep in grpn])
    lesgroepen: list[Lesgroep] = []
    for groep, lesdata in datalist:
        if lesdata:
            groepdata = get_info(*lesdata, lln=lln, docs=docs)
            lesgroep = Lesgroep(vak, groep, lj, *groepdata)
            lesgroepen.append(lesgroep)
    return (vak, lesgroepen)
